#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/sysctl.h>
#include <sys/systm.h>

#include <machine/clock.h>

extern void frame_push_pop(u_long iterations);
extern void frame_mov_mov(u_long iterations);

SYSCTL_NODE(_debug, OID_AUTO, pvm, CTLFLAG_RW, 0, "");

static u_long iterations = 1000000;
SYSCTL_ULONG(_debug_pvm, OID_AUTO, iterations, CTLFLAG_RW, &iterations, 0, 0);

static u_long usecs;
SYSCTL_ULONG(_debug_pvm, OID_AUTO, usecs, CTLFLAG_RW, &usecs, 0, 0);

static uint64_t
read_tsc(void)
{
	uint64_t tsc;
	u_int regs[4];

	do_cpuid(0, regs);
	tsc = rdtsc();
	do_cpuid(0, regs);

	return (tsc);
}

static int
sysctl_doit(SYSCTL_HANDLER_ARGS)
{
	int error, doit;
	uint64_t tsc1, tsc2, pushit, rflags;

	doit = 0;
	error = sysctl_handle_int(oidp, &doit, 0, req);
	if (error || req->newptr == NULL || doit == 0)
		return (error);

	switch (doit) {
	case 1:
		pushit = 1;
		break;
	case 2:
		pushit = 0;
		break;
	default:
		error = EINVAL;
		break;
	}

	if (error)
		return (error);

	rflags = intr_disable();
	tsc1 = read_tsc();

	if (pushit)
		frame_push_pop(iterations);
	else
		frame_mov_mov(iterations);

	tsc2 = read_tsc();
	intr_restore(rflags);

	usecs = ((tsc2 - tsc1) * 1000000) / tsc_freq;

	return (0);
}

SYSCTL_PROC(_debug_pvm, OID_AUTO, doit, CTLTYPE_INT|CTLFLAG_RW, 0, 0, 
    sysctl_doit, "I", "");

static int
pvm_modevent(module_t mod, int type, void *unused)
{

        switch (type) {
        case MOD_LOAD:
		break;
	case MOD_UNLOAD:
		break;
	default:
		return (EINVAL);
	}

	return (0);
}

static moduledata_t pvm_mod_data= {
	"pvm",
	pvm_modevent,
};

DECLARE_MODULE(test, pvm_mod_data, SI_SUB_SMP + 1, SI_ORDER_ANY);
